/*******************************************************************************

  Intel SmartPHY DSL PCIe TC driver
         Copyright (c) 2020, MaxLinear, Inc.
         Copyright 2016 - 2020 Intel Corporation
         Copyright 2015 - 2016 Lantiq Beteiligungs-GmbH & Co. KG
         Copyright 2009 - 2014 Lantiq Deutschland GmbH
         Copyright 2007 - 2008 Infineon Technologies AG

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

*******************************************************************************/
#define DEBUG
#define pr_fmt(fmt) KBUILD_MODNAME ":%s: " fmt, __func__

#ifndef TC_COMMIT_ID
#define TC_COMMIT_ID Not_Set
#endif
#ifndef TC_BUILD_DATE
#define TC_BUILD_DATE Not_Set
#endif
#ifndef TC_BRANCH
#define TC_BRANCH Not_Set
#endif
#ifndef TC_BUILD_PATH
#define TC_BUILD_PATH Not_Set
#endif

#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/atomic.h>
#include <linux/uaccess.h>
#include <linux/bitops.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/interrupt.h>
#include <linux/dma-mapping.h>
#include <linux/wait.h>
#include <linux/seq_file.h>
#include <linux/printk.h>
#include <linux/etherdevice.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/stringify.h>

#include "inc/tc_main.h"
#include "inc/platform.h"
#include "inc/reg_addr.h"
#include "inc/tc_common.h"
#include "inc/tc_api.h"
#include "inc/tc_proc.h"

/*
 * @History:
 * Version 1.5.7:
 *   FEATURE_CONF_DESC_LENGTH
 *     518 TC driver supports configuration
 *     number of US descriptors.
 *     Modification: ACA, US FP, US QoS, OutQ blocks.
 * Version 1.5.8:
 *  Support power saving and PPE dynamic frequency setting
 * Version 1.5.9:
 * 	FEATURE_POWER_DOWN
 *
 * Version 1.5.12.9:
 *      Fix italy-ZTE Bonding Line1 traffic issue (IPSUP-54539)
 * Version 1.5.12.10:
 *      Root cause:  After PCIe RC and EP driver were removed,
 *		there was still traffic access across PCIe which was
 *		causing the bus hang.
 *		Fix: Added remove functionality in TC driver during system
 *		reboot process.
 * Version 1.5.12.11:
 * 	Fix for heartbeat callback issue (UGW_SW-41874).
 * Version 1.5.12.12:
 * 	Fix for ifconfig ptm0 stats (UGW_SW-49080).
 * Version 1.6.1.0:
 * 	Feature ATMBOND for Century Link.
 * Version 1.6.1.1:
 *      ATM Bonding: ATM single line -HW Acceleration support
 *      https://jira-chd.intel.com/browse/UGW_SW-55264
 * Version 1.6.1.2:
 *	ATM Bonding: Mode change recover if nas exists
 * Version 1.6.1.3:
 *      ATM Bonding: QoS drop dis, fake tc bit set, tc feature set
 * Version 1.6.1.5:
 *	ATMBOND: ugw 8.5.0 GA code freeze
 * Version 1.6.1.7:
 *	ATMBOND: DS FP code, couple of bug fixes.
 * Version 1.6.1.8:
 *	ATMBOND: RX skb tail fix (UGW_SW-56909).
 * Version 1.6.1.9:
 *	ATMBOND: DP dev null check remove (UGW_SW-56626).
 * Version 1.6.2.0:
 *	PTM: UGW_SW-56626 PTM rx pkt before showtime enter crash fix.
 * Version 1.6.2.1:
 *      ATMBOND: Reduce codeword /local queue buffer size
 *
 * Version 1.6.2.2:
 *	Call trace in ptm_rx() while switching from ATM to PTM mode 
 *	https://iljira.phoenix.local/browse/UGW_SW-59776
 *
 * Version 1.6.2.2:
 *	Call trace in TC driver during force reboot 
 *	https://iljira.phoenix.local/browse/UGW_SW-58352
**/
#define DRV_VERSION	"1.6.2.3"

#define TCDRV_FEATURE_VAL_FOR_DSL 1

static const char tc_drv_name[] = "vrx518_tc";
static const char tc_drv_ver[] = DRV_VERSION;
static const char tc_drv_string[] =
			"Intel(R) SmartPHY DSL(VRX518) PCIe TC Driver";
static const char tc_drv_copyright[] =
			"Copyright (c) 2020 MaxLinear, Inc.";
static struct tc_priv *g_tc_priv;
static int p2p_en = 1;
static int loopback_en;
static int sharing_cdma_en = 1;
static u32 umt_period = UMT_DEF_PERIOD;
static u32 cdma_desc_loc = 1;
/* 2017-12-15: from version 1.5.5, turn on power saving mode */
static u32 powersaving = 1;
#ifdef FIX_DMA_BYTE_ALIGNMENT
static u32 ena_dma_byte_alignment;
#endif
#ifdef FEATURE_PPE_HEARTBEAT
/* change it to 0 then default ppe heartbeat is disable*/
#define DEFAULT_PPE_HEARTBEAT_ENABLE	1
/* default timer is 100 ms */
#define DEFAULT_PPE_HEARTBEAT_TIME		100
static struct timer_list g_ppe_heartbeat_timer;

int ppe_close_pend = 0;

#endif
module_param(p2p_en, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(p2p_en, "PTM Bonding directly access Peer Enable flag.");
module_param(loopback_en, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(loopback_en, "Enable-1/Disable-0 the DFE loopback mode.");
module_param(sharing_cdma_en, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(sharing_cdma_en, "Enable-1/Disable-0 the sharing CDMA mode.");
module_param(umt_period, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(umt_period, "Adjust the frequency of the UMT message, default: 200us");
module_param(cdma_desc_loc, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(cdma_desc_loc, "CDMA descriptor address: 0 - SRAM/1 - FPI");
module_param(powersaving, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(powersaving, "Powersaving configuration: 0 - Disable/1-Enable");

#ifdef FIX_DMA_BYTE_ALIGNMENT
module_param(ena_dma_byte_alignment, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(ena_dma_byte_alignment, "Enable DMA Byte Alignment SG: 0 - Disable/1-Enable");
#endif

u32 read_a_mem_value(char *s, int ep, u32 reg)
{
	struct tc_priv *priv = g_tc_priv;
	u32 val, membase;

	if (!priv) {
		printk("tc priv is null\n");
		return -1;
	}
	membase = (u32)(priv->ep_dev[ep].membase);
	val = readl(membase + reg);
	//printk("%s =================>line-%d 0x%x\n", s, ep, val);
	return val;
}

static int set_feature_set_val_for_dsl(u8 set)
{
	struct tc_priv *priv = g_tc_priv;
	u32 membase, rd, val = 0;
	int i;

	if (!priv && ((priv->ep_num < 0)||(priv->ep_num > 2)))
		return -1;
	for (i = 0; i < priv->ep_num; i++) {
		membase = (u32)(priv->ep_dev[i].membase);
		rd = val = readl(membase + DSL_FW_TC_VER_INDC);
		pr_cont("=====================@0x%x DSL feature set val old: 0x%x", (membase + DSL_FW_TC_VER_INDC), val);
		if (set) {
			val = 0x5a5a0000;
			val = val | (u16)TCDRV_FEATURE_VAL_FOR_DSL;
		} else {
			val = 0;
		}
		pr_cont(" new: 0x%x\n", val);
		if (rd != val)
			writel(val, membase + DSL_FW_TC_VER_INDC);
	}
	return 0;
}

static int pcie_ep_probe(struct tc_priv *priv)
{
	int dev_num;
	int i;

	if (dc_ep_dev_num_get(&dev_num)) {
		tc_err(priv, MSG_INIT, "Failed to get total PCIe device number\n");
		return -EIO;
	}
	tc_dbg(priv, MSG_INIT, "Total %d VRX518 EP detected\n", dev_num);

	for (i = 0; i < dev_num; i++) {
		if (dc_ep_dev_info_req(i, DC_EP_INT_PPE, &priv->ep_dev[i])) {
			tc_err(priv, MSG_INIT, "failed to get pcie ep %d\n", i);
			goto err1;
		}
        /* 20180921: Turn on PMU_EMA to prevent BAR hang in DSL FW */
        tc_clkon(&priv->ep_dev[i], PMU_EMA);
	}

	priv->ep_num = i;
	priv->switch_exist = priv->ep_dev[0].switch_attached;

	return 0;
err1:
	for (i = i - 1; i >= 0; i--)
		dc_ep_dev_info_release(i);

	return -EIO;
}

static void pcie_ep_release(struct tc_priv *priv)
{
	int i;

	for (i = 0; i < priv->ep_num; i++) {
        tc_clkoff(&priv->ep_dev[i], PMU_EMA);
		dc_ep_dev_info_release(i);
    }
}

static void tc_def_recv(struct net_device *pdev, struct sk_buff *skb)
{
	tc_dbg(g_tc_priv, MSG_RX, "TC is not loaded\n");
	dev_kfree_skb_any(skb);
}

static void tc_def_irq_on(u32 irq_no)
{
	return;
}

static void tc_def_irq_off(u32 irq_no)
{
	return;
}

static int tc_def_getqid(struct net_device *pdev, struct sk_buff *skb,
		void *vcc, uint32_t flags)
{
	return -1;
}

static inline void tc_def_ops_setup(struct tc_priv *priv)
{
	spin_lock_bh(&priv->tc_lock);
	priv->tc_ops.recv = tc_def_recv;
	priv->tc_ops.irq_on = tc_def_irq_on;
	priv->tc_ops.irq_off = tc_def_irq_off;
	priv->tc_ops.get_qid = tc_def_getqid;
	priv->tc_ops.showtime_enter = NULL;
	priv->tc_ops.showtime_exit = NULL;
	/*priv->tc_ops.tc_reset = NULL;*/
	spin_unlock_bh(&priv->tc_lock);
}

static inline void init_local_param(struct tc_priv *priv)
{
	if (priv->ep_num > 1)
		priv->param.bonding_en = 1;
	priv->param.dfe_loopback = loopback_en;
	priv->param.p2p = p2p_en;
	priv->param.umt_period = umt_period;
	priv->param.sharing_cdma_en = sharing_cdma_en;
	priv->param.cdma_desc_loc = cdma_desc_loc;
	priv->param.ps = powersaving;
	#ifdef FIX_DMA_BYTE_ALIGNMENT
	priv->param.ena_dma_tc_sg = ena_dma_byte_alignment;
	#endif
#ifdef FEATURE_CONF_DESC_LENGTH
	priv->param.conf_us_fp_desq_len = 32;
	priv->param.conf_us_qos_queue_len = 32;
	priv->param.conf_us_outq_len = 32;
	priv->param.conf_us_local_q0_desq_len = 16;
#endif
	priv->tc_mode = TC_NONE_MODE;
	priv->tc_stat = NO_TC;
	priv->tc_idx = -1;
	spin_lock_init(&priv->tc_lock);
	spin_lock_init(&priv->tc_req_lock);
	tc_def_ops_setup(priv);
}

void tc_unload(struct tc_priv *priv)
{
printk("%s, ()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()\n", __func__);
	if (priv == NULL) {
		pr_err("%s-%d: Invalid priv\n", __func__, __LINE__);
        return;
    }

	if (priv->tc_mode == TC_NONE_MODE)
		return;
	tc_def_ops_setup(priv);
	if (priv->tc_mode == TC_PTM_SL_MODE ||
		priv->tc_mode == TC_PTM_BND_MODE)
		ptm_tc_unload(priv->tc_mode);
	else if (priv->tc_mode == TC_ATM_SL_MODE ||
			priv->tc_mode == TC_ATM_BND_MODE) {
		atm_tc_unload(priv->tc_mode);
	} else {
		pr_err("%s-%d: Invalid TC MODE %d\n",
			__func__, __LINE__, priv->tc_mode);
		return;
	}

#ifdef FEATURE_PPE_HEARTBEAT
	if (g_ppe_heartbeat_timer.entry.pprev != NULL)
		del_timer_sync(&g_ppe_heartbeat_timer);
#endif
	set_feature_set_val_for_dsl(0);
	spin_lock_bh(&priv->tc_lock);
	priv->tc_mode = TC_NONE_MODE;
	priv->tc_stat = NO_TC;
	priv->tc_idx = -1;
	priv->priv = NULL;
	spin_unlock_bh(&priv->tc_lock);
}

int is_nas_iface_exist(void)
{
	int i;
	struct net_device *dev = NULL;
	char name[IFNAMSIZ] = {"nas"};

	for (i=0; i<32; i++) {
		snprintf(name+ 3, 13, "%d", i);
		if ((dev = dev_get_by_name (&init_net, name)) != NULL) {
			dev_put(dev);
			return 1;
		}
	}
	return 0;
}

void tc_load(struct tc_priv *priv, u32 id, enum dsl_tc_mode mode)
{
printk("%s, ()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()\n", __func__);
    if (priv == NULL) {
        pr_err("%s-%d: Invalid priv\n", __func__, __LINE__);
        return;
    }

    if (mode == TC_NONE_MODE) {
#ifdef FEATURE_POWER_DOWN
	    if (priv->dsl_pwr_policy[id]) {
		power_configure(priv, priv->dsl_pwr_policy[id] - 1,
		    id, SW_OFF_VRX518_PLL);
		tc_dbg(priv, MSG_SWITCH, "TC Request OFF on line [%d] with PLL [%d]\n",
		    id, priv->dsl_pwr_policy[id]);
		/* clear the policy */
		priv->dsl_pwr_policy[id] = 0;
	    }
#endif
		priv->tc_stat = NO_TC;
		return;
	} else if (is_nas_iface_exist()) {
		printk("NAS iface is not un-registered; delaying TC Load ...\n");
		return;
	}

	tc_dbg(priv, MSG_SWITCH, "TC Load on [%d]: from[%d] to[%d]\n",
		id, priv->tc_mode,
		mode);

	spin_lock_bh(&priv->tc_lock);
	priv->tc_mode = mode;
	spin_unlock_bh(&priv->tc_lock);
	if (mode == TC_PTM_SL_MODE || mode == TC_PTM_BND_MODE)
		ptm_tc_load(priv, id, mode);
	else if (mode == TC_ATM_SL_MODE || mode == TC_ATM_BND_MODE)
		atm_tc_load(priv, id, mode);
#ifdef FEATURE_PPE_HEARTBEAT
	u32 membase, val;
	int i;
    for (i = 0; i < priv->ep_num; i++) {
        membase = (u32)(priv->ep_dev[i].membase);
        val = readl(membase + DSL_TC_INDICATE);
	pr_cont("=====================heartbeat set val old: 0x%x", val);
        val |= INDICATE_PPE_HEARTBEAT;
	pr_cont(" new: 0x%x\n", val);
        writel(val, membase + DSL_TC_INDICATE);
    }
	/* set timer with ms */
	/* if timer pprev is NULL, means timer expired and need add timer again */
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,8,0)
		g_ppe_heartbeat_timer.expires = jiffies + (priv->tc_ppe_heartbeat_time * HZ / 1000);
		add_timer(&g_ppe_heartbeat_timer);
#else
     if (g_ppe_heartbeat_timer.entry.pprev == NULL) {
         g_ppe_heartbeat_timer.expires = jiffies + (priv->tc_ppe_heartbeat_time * HZ / 1000);
         add_timer(&g_ppe_heartbeat_timer);
     } else {
         tc_err(priv, MSG_INIT, "timer error! pprev=0x%x\n", g_ppe_heartbeat_timer.entry.pprev);
     }
#endif
#endif
	set_feature_set_val_for_dsl(1);
	spin_lock_bh(&priv->tc_lock);
	priv->tc_stat = TC_RUN;
	priv->tc_idx = id;
	spin_unlock_bh(&priv->tc_lock);
}

void tc_request(u32 id, enum dsl_tc_mode tc_mode)
{
	struct tc_priv *priv = g_tc_priv;
printk("%s, ()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()() idx: %d\n", __func__, id);
	tc_unload(priv);
	tc_load(priv, id, tc_mode);
}

void tc_get_drv_version(const char **drv_name, const char **drv_ver)
{
	*drv_name = tc_drv_name;
	*drv_ver = tc_drv_ver;
}
#ifdef FEATURE_PPE_HEARTBEAT

void tc_ppe_heartbeat_callback(unsigned long dummy)
{
	struct tc_priv *priv = g_tc_priv;
	int i;
	u32 mbox_ier;
	u32 membase;

	/* write to mailbox to ppe heartbeat */
	/* should wait ppe sync then start ppe heartbeat */
	for (i = 0; i < priv->ep_num; i++) {
		membase = (u32)(priv->ep_dev[i].membase);
		if (priv->priv == NULL)
		return;
		if ((priv->tc_ppe_heartbeat[i].ppe_heartbeat_enable) &&
			((readl(membase + DSL_TC_INDICATE) & 0x07) != 0)) {
			mbox_ier = readl(membase + MBOX_IGU_IER(MBOX_IGU1));
			mbox_ier |= MBOX_PKT_RX;
			writel(mbox_ier, membase + MBOX_IGU_IER(MBOX_IGU1));
			readl(membase + MBOX_IGU_IER(MBOX_IGU1));

			writel(MBOX_FW_WAKEUP, membase + MBOX_IGU_ISRS(PPE_FW_MBOX0));
		}
	}

	/* set timer with ms */
	/* if timer pprev is NULL, means timer expired and need add timer again */
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,8,0)
		g_ppe_heartbeat_timer.expires = jiffies + (priv->tc_ppe_heartbeat_time * HZ / 1000);
		add_timer(&g_ppe_heartbeat_timer);
#else
     if (g_ppe_heartbeat_timer.entry.pprev == NULL) {
         g_ppe_heartbeat_timer.expires = jiffies + (priv->tc_ppe_heartbeat_time * HZ / 1000);
         add_timer(&g_ppe_heartbeat_timer);
     } else {
         tc_err(priv, MSG_INIT, "timer error!\n");
         tc_info(priv, MSG_INIT, "entry=0x%x\n", g_ppe_heartbeat_timer.entry);
         tc_info(priv, MSG_INIT, "pprev=0x%x\n", g_ppe_heartbeat_timer.entry.pprev);
         tc_info(priv, MSG_INIT, "flags=0x%x\n", g_ppe_heartbeat_timer.flags);
         tc_info(priv, MSG_INIT, "data=%p\n", g_ppe_heartbeat_timer.data);
         tc_info(priv, MSG_INIT, "expires=%p\n", g_ppe_heartbeat_timer.expires);
     }
#endif
}
#endif

static int pre_exit_cleanup(void)
{
	struct tc_priv *priv;
	priv = g_tc_priv;
	platform_dsl_exit();
	spin_lock_bh(&priv->tc_lock);
	if (priv->tc_stat == TC_INIT || priv->tc_stat == TC_SWITCHING)
		goto tc_error;
	spin_unlock_bh(&priv->tc_lock);
	if (priv->tc_stat == TC_RUN)
		tc_unload(priv);

	ptm_exit();
	atm_exit();
	ppe_fw_unload(priv);
	tc_proc_exit(priv);
	platform_exit();
	pcie_ep_release(priv);
	kfree(priv);
	tc_gentlk_exit();

#ifdef FEATURE_PPE_HEARTBEAT
	if (g_ppe_heartbeat_timer.entry.pprev != NULL)
		del_timer_sync(&g_ppe_heartbeat_timer);
#endif
	return 0;
tc_error:
	spin_unlock_bh(&priv->tc_lock);
	tc_dbg(priv, MSG_SWITCH,
		"TC is not ready to exit: %d\n", priv->tc_stat);
	return -1;
}

static int smartphy_reboot_notifier(struct notifier_block *nb,
				unsigned long action, void *unused)
{       
 	printk("reboot action is %d \n",action);
	
	/*force unload ATM TC */
	ppe_close_pend =0;
	
	if (!pre_exit_cleanup())
		return NOTIFY_DONE;
	else
		return NOTIFY_STOP;
}

static struct notifier_block smartphy_reboot_nb = {
		.notifier_call = smartphy_reboot_notifier,
};

static int __init tc_drv_init(void)
{
	struct tc_priv *priv;
	int ret;
	size_t size;
	size = ALIGN(sizeof(struct tc_priv), TCPRIV_ALIGN) + plat_priv_sz();
	priv = kzalloc(size, GFP_KERNEL);
	if (!priv) {
		pr_err("Alloc tc priv fail!\n");
		return -ENOMEM;
	}
	g_tc_priv = priv;

	priv->msg_enable = TC_DEF_DBG;
	ret = pcie_ep_probe(priv);
	if (ret < 0) {
		tc_err(priv, MSG_INIT, "ep probe fail!\n");
		goto err1;
	}

	init_local_param(priv);

	ret = platform_init(priv, tc_drv_name);
	if (ret < 0) {
		tc_err(priv, MSG_INIT, "soc platform init fail!\n");
		goto err2;
	}
	ret = ppe_fw_load(priv);
	if (ret < 0)
		goto err3;

	tc_gentlk_init(priv);
	tc_proc_init(priv);

	ret = register_reboot_notifier(&smartphy_reboot_nb);
#ifdef FEATURE_PPE_HEARTBEAT
	int i;
	init_timer(&g_ppe_heartbeat_timer);
	g_ppe_heartbeat_timer.function = tc_ppe_heartbeat_callback;
	for (i = 0; i < priv->ep_num;i++) {
		priv->tc_ppe_heartbeat[i].ppe_heartbeat_enable = DEFAULT_PPE_HEARTBEAT_ENABLE;
		priv->tc_ppe_heartbeat[i].ppe_heartbeat_counter = 0;
	}
	priv->tc_ppe_heartbeat_time = DEFAULT_PPE_HEARTBEAT_TIME;
#endif
	tc_info(priv, MSG_INIT, "%s - version %s\n",
		tc_drv_string, tc_drv_ver);
	tc_info(priv, MSG_INIT, "%s\n", tc_drv_copyright);

	return 0;
err3:
	platform_exit();
err2:
	pcie_ep_release(priv);
err1:
	kfree(priv);
	g_tc_priv = NULL;
	return ret;
}
module_init(tc_drv_init);

/**
 * 1. Unregister DSL callback function
 * 2. Unload TC
 * 3. free ATM/PTM/Bonding driver resources
 * 4. free SoC resource
 * 5. free proc
 * 6. free tc_priv
 * 7. free netlink resource
 */
static void __exit tc_drv_exit(void)
{
	if (!pre_exit_cleanup())
		pr_info("%s driver unloaded\n", tc_drv_name);
	return;
}
module_exit(tc_drv_exit);


MODULE_AUTHOR("Intel Corporation, <yixin.zhu@intel.com>");
MODULE_DESCRIPTION("Intel(R) SmartPHY PCIe EP/ACA Driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);
MODULE_INFO(version, __stringify(DRV_VERSION));
MODULE_INFO(branch, __stringify(TC_BRANCH));
MODULE_INFO(commitid, __stringify(TC_COMMIT_ID));
MODULE_INFO(buildpath, __stringify(TC_BUILD_PATH));
MODULE_INFO(builddate, __stringify(TC_BUILD_DATE));
